#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2021-03-14 15:54:40
# @Author  : Chenghao Mou (mouchenghao@gmail.com)

import os
import logging.config
from typing import List

logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": True,
    }
)

from strsimpy.normalized_levenshtein import NormalizedLevenshtein
from touchbar_lyric import Song
from touchbar_lyric.service.netease import netease_music_search
from touchbar_lyric.service.qq import qq_music_search
from diskcache import FanoutCache

CACHE = os.path.join(os.path.expanduser("~"), ".cache")

if not os.path.exists(CACHE):
    os.mkdir(CACHE)

cache = FanoutCache(CACHE, timeout=2)


@cache.memoize(typed=True, expire=None, tag="lyric")
def universal_search(title: str, artists: str) -> List[Song]:  # pragma: no cover
    songs = []
    songs.extend(qq_music_search(title, artists))
    songs.extend(netease_music_search(title, artists))

    songs = sorted(
        songs,
        key=lambda s: (
            NormalizedLevenshtein().similarity(s.title, s.target_title)
            + NormalizedLevenshtein().similarity(s.artists, s.target_artists),
        ),
        reverse=True,
    )

    return songs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2021-03-01 19:51:54
# @Author  : Chenghao Mou (mouchenghao@gmail.com)

import bisect
from collections import namedtuple
from typing import Optional, List

import applescript
from loguru import logger

MediaInformation = namedtuple(
    "MediaInformation", ["name", "artists", "position", "state", "durantion"]
)


def get_info(app: str) -> Optional[MediaInformation]:
    """Get media information with apple script.

    Parameters
    ----------
    app : str
        The name of the application

    Returns
    -------
    Optional[MediaInformation]
        MediaInformation object

    Examples
    --------
    >>> ans = get_info("Spotify")
    >>> assert ans is None or isinstance(ans, MediaInformation)
    """

    script: str = f"""
    on run
        if application "{app}" is running then
            tell application "{app}"
                set currentInfo to {{name of current track, "|", artist of current track, "|", player position, "|", player state, "|", duration of current track}}
            end tell
        else
            set currentInfo to "Empty"
        end if
        return currentInfo
    end run
    """

    r = applescript.run(script)

    logger.debug(r.out)

    ans: Optional[MediaInformation] = None
    if r.code == 0 and r.out != "Empty":
        segments = r.out.split(", ")
        ans = MediaInformation(
            segments[0],
            segments[2],
            float(segments[4]),
            {"playing": 2, "paused": 1, "stopped": 0}.get(segments[6], 0),
            float(segments[8]) // 1000
            if "." not in segments[8]
            else float(segments[8]),
        )

    logger.debug(ans)

    return ans


def search_intervals(
    intervals: List[float], position: float
) -> int:  # pragma: no cover
    """Search a timestamp in a list of intervals.

    Parameters
    ----------
    intervals : List[float]
        List of timestamps
    position : float
        Current timestamp

    Returns
    -------
    int
        Index of the interval

    Examples
    --------
    >>> search_intervals([12, 15], 13)
    0
    >>> search_intervals([12, 15], 7)
    0
    >>> search_intervals([12, 15], 16)
    1
    """
    idx = max(0, bisect.bisect_left(intervals, position) - 1)

    if len(intervals) > idx >= 0 and (
        idx == 0
        or idx == len(intervals) - 1
        or (intervals[idx] <= position <= intervals[idx + 1])
    ):
        return idx

    return -1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2020-03-10 10:54:37
# @Author  : Chenghao Mou (mouchenghao@gmail.com)

import base64
import binascii
import json
import os
from typing import Any, Dict, List

import requests

from Crypto.Cipher import AES
from loguru import logger

from touchbar_lyric import Song


class NeteaseRequest:  # pragma: no cover
    """A request wrapper for Netease music."""

    session = requests.Session()
    session.headers.update(
        {
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "Accept-Charset": "UTF-8,*;q=0.5",
            "Accept-Encoding": "gzip,deflate,sdch",
            "Accept-Language": "en-US,en;q=0.8",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:60.0) Gecko/20100101 Firefox/60.0",
            "referer": "http://music.163.com/",
        }
    )

    @classmethod
    def encode_netease_data(cls, data) -> str:  # pragma: no cover
        data = json.dumps(data)
        key = binascii.unhexlify("7246674226682325323F5E6544673A51")
        encryptor = AES.new(key, AES.MODE_ECB)
        pad = 16 - len(data) % 16
        fix = chr(pad) * pad
        byte_data = (data + fix).encode("utf-8")
        return binascii.hexlify(encryptor.encrypt(byte_data)).upper().decode()

    @classmethod
    def encrypted_request(cls, data) -> dict:  # pragma: no cover
        MODULUS = (
            "00e0b509f6259df8642dbc35662901477df22677ec152b5ff68ace615bb7"
            "b725152b3ab17a876aea8a5aa76d2e417629ec4ee341f56135fccf695280"
            "104e0312ecbda92557c93870114af6c9d05c4f7f0c3685b7a46bee255932"
            "575cce10b424d813cfe4875d3e82047b97ddef52741d546b8e289dc6935b"
            "3ece0462db0a22b8e7"
        )
        PUBKEY = "010001"
        NONCE = b"0CoJUm6Qyw8W8jud"
        data = json.dumps(data).encode("utf-8")
        secret = cls.create_key(16)
        params = cls.aes(cls.aes(data, NONCE), secret)
        encseckey = cls.rsa(secret, PUBKEY, MODULUS)
        return {"params": params, "encSecKey": encseckey}

    @classmethod
    def aes(cls, text, key):  # pragma: no cover
        pad = 16 - len(text) % 16
        text = text + bytearray([pad] * pad)
        encryptor = AES.new(key, 2, b"0102030405060708")
        ciphertext = encryptor.encrypt(text)
        return base64.b64encode(ciphertext)

    @classmethod
    def rsa(cls, text, pubkey, modulus):  # pragma: no cover
        text = text[::-1]
        rs = pow(int(binascii.hexlify(text), 16), int(pubkey, 16), int(modulus, 16))
        return format(rs, "x").zfill(256)

    @classmethod
    def create_key(cls, size):  # pragma: no cover
        return binascii.hexlify(os.urandom(size))[:16]

    @classmethod
    def request(
        cls, url: str, data: Dict[str, Any], method: str = "POST"
    ) -> Dict[str, Any]:

        results = {}
        status = requests.codes.ok
        text = ""

        try:
            if method == "GET":
                resp = cls.session.get(url, params=data, timeout=20)
            else:
                resp = cls.session.post(url, data=data, timeout=20)
            results = resp.json()
            text = resp.text
            status = resp.status_code
        except Exception as e:
            results = {}
            logger.debug(e)

        if status != requests.codes.ok or not text:
            results = {}

        return results


def netease_music_get_lyric(idx) -> str:
    data = NeteaseRequest.encrypted_request(
        {"csrf_token": "", "id": idx, "lv": -1, "tv": -1}
    )
    return (
        NeteaseRequest.request(
            url="https://music.163.com/weapi/song/lyric", method="POST", data=data
        )
        .get("lrc", {})
        .get("lyric", "")
    )


def netease_music_search(title: str, artists: str) -> List[Song]:
    """
    Search from Netease Music with artists and title.
    Parameters
    ----------
    title : str
        Name of the song
    artists : str
        Names of the artists
    Returns
    -------
    List[Song]
        List of songs
    Examples
    --------
    >>> songs = netease_music_search("海阔天空", "Beyond")
    >>> len(songs) > 0
    True
    >>> any(s.anchor(10) is not None for s in songs)
    True
    """

    eparams = {
        "method": "POST",
        "url": "http://music.163.com/api/cloudsearch/pc",
        "params": {"s": title, "type": 1, "offset": 0, "limit": 30},
    }
    data = {"eparams": NeteaseRequest.encode_netease_data(eparams)}

    res_data = (
        NeteaseRequest.request(
            "http://music.163.com/api/linux/forward", method="POST", data=data
        )
        .get("result", {})
        .get("songs", [])
    )
    songs = []
    for _, item in enumerate(res_data[:2]):
        if item.get("id", None) is not None:
            s = Song(
                title=item.get("name", ""),
                artists=",".join(
                    [x["name"] for x in item.get("ar", []) if "name" in x]
                ),
                target_title=title,
                target_artists=artists,
                lyric=netease_music_get_lyric(idx=item["id"]),
            )
            songs.append(s)
    return songs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2021-03-14 14:37:35
# @Author  : Chenghao Mou (mouchenghao@gmail.com)

import html

import regex as re
from typing import Dict, Optional, Tuple, List
from dataclasses import dataclass

from touchbar_lyric.utility import search_intervals


@dataclass
class Song:

    title: str
    artists: str
    target_title: str
    target_artists: str
    lyric: str
    lines: Optional[Dict[Tuple[float, float], str]] = None
    intervals: Optional[List[Tuple[float, float]]] = None

    def __post_init__(self):

        lyric = self.lyric
        lyric = html.unescape(lyric)

        self.lines = []
        self.intervals = []
        for line in lyric.split("\n"):
            info, *words = line.rsplit("]", 1)
            timestamp = re.search(r"\[([0-9]+):([0-9]+)\.([0-9]+)\]", info + "]")
            if not timestamp:
                continue
            minute, second, subsecond = (
                timestamp.group(1),
                timestamp.group(2),
                timestamp.group(3),
            )
            curr_stamp = int(minute) * 60 + int(second) + int(subsecond) / 100
            self.lines.append((curr_stamp, "".join(words)))
            self.intervals.append(curr_stamp)

    def anchor(self, timestamp: float) -> Optional[str]:
        """Find current timestamp for this song.

        Parameters
        ----------
        timestamp : float
            Current timestamp

        Returns
        -------
        Optional[str]
            A line or None

        Examples
        --------
        >>> song = Song("Hello", "Adele", "Hello", "Adele", "[01:12.34]Hello")
        >>> song.anchor(60)
        'Hello'
        >>> song.anchor(120)
        'Hello'
        >>> song = Song("Hello", "Adele", "Hello", "Adele", "")
        >>> song.anchor(10) is None
        True
        """
        if not self.intervals:
            return None

        idx = search_intervals(self.intervals, timestamp)
        if idx != -1:
            return self.lines[idx][-1]

        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2020-08-16 09:10:24
# @Author  : Chenghao Mou (mouchenghao@gmail.com)


from typing import List

from QQMusicAPI import QQMusic

from touchbar_lyric import Song


def qq_music_search(title: str, artists: str) -> List[Song]:
    """
    Search from QQ Music with artists and title.
    Parameters
    ----------
    title : str
        Name of the song
    artists : str
        Names of the artists
    Returns
    -------
    List[Song]
        List of songs
    Examples
    --------
    >>> songs = qq_music_search("海阔天空", "Beyond", ignore_cache=True)
    >>> len(songs) > 0
    True
    >>> any(s.anchor(10) is not None for s in songs)
    True
    """
    response = QQMusic.search(title)
    songs = []
    top: int = 3
    for song in response.data:
        lyric = song.lyric
        lyric.extract()
        if lyric.lyric or lyric.trans:
            content = lyric.lyric or lyric.trans
            songs.append(
                Song(
                    title=song.name,
                    artists=",".join([x.name for x in song.singer]),
                    target_title=title,
                    target_artists=artists,
                    lyric=content,
                )
            )
            top -= 1
            if top == 0:
                break
    return songs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2021-03-14 15:29:12
# @Author  : Chenghao Mou (chenghao@gmail.com)

import typer
import logging.config

from hanziconv import HanziConv
from loguru import logger
from touchbar_lyric.utility import get_info
from touchbar_lyric.service import universal_search

logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": True,
    }
)


def run(
    app: str = typer.Option(default="Spotify", help="Application to track"),
    debug: bool = typer.Option(
        default=False, is_flag=True, help="To show debug messages or not"
    ),
    traditional: bool = typer.Option(
        default=False,
        is_flag=True,
        help="Translate lyrics into Traditional Chinese if possible",
    ),
):  # pragma: no cover
    {True: logger.enable, False: logger.disable}[debug]("touchbar_lyric")

    if not debug:
        logger.disable("touchbar_lyric")
        logger.disable("__main__")

    media_info = get_info(app)
    if media_info is None:
        return

    songs = universal_search(media_info.name, media_info.artists)

    for song in songs:
        if song.anchor(media_info.position):
            line: str = song.anchor(media_info.position)
            if traditional:
                line = HanziConv.toTraditional(line)
            print(line)
            break


if __name__ == "__main__":  # pragma: no cover
    typer.run(run)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['touchbar_lyric', 'touchbar_lyric.service', 'touchbar_lyric.utility']

package_data = \
{'': ['*']}

install_requires = \
['applescript>=2021.2.9,<2022.0.0',
 'cachier>=1.5.0,<2.0.0',
 'coverage>=5.5,<6.0',
 'diskcache>=5.2.1,<6.0.0',
 'hanziconv>=0.3.2,<0.4.0',
 'loguru>=0.5.3,<0.6.0',
 'pycrypto>=2.6.1,<3.0.0',
 'qqmusic-api>=0.1,<0.2',
 'regex>=2021.3.17,<2022.0.0',
 'strsimpy>=0.2.0,<0.3.0',
 'typer>=0.3.2,<0.4.0']

setup_kwargs = {
    'name': 'touchbar-lyric',
    'version': '0.6.8',
    'description': 'Display lyrics on your touchbar with BTT',
    'long_description': '<center><h1>Synced Lyric on TouchBar</h1></center>\n\n[![Codacy Badge](https://api.codacy.com/project/badge/Grade/77de523131f9441997db18c608b3c54e)](https://app.codacy.com/manual/mouchenghao/touchbar-lyric?utm_source=github.com&utm_medium=referral&utm_content=ChenghaoMou/touchbar-lyric&utm_campaign=Badge_Grade_Dashboard) [![Build Status](https://travis-ci.com/ChenghaoMou/touchbar-lyric.svg?branch=master)](https://travis-ci.com/ChenghaoMou/touchbar-lyric) [![Codacy Badge](https://app.codacy.com/project/badge/Coverage/aadeca6117a14aa6b655e21d5bbc09ea)](https://www.codacy.com/manual/mouchenghao/touchbar-lyric?utm_source=github.com&utm_medium=referral&utm_content=ChenghaoMou/touchbar-lyric&utm_campaign=Badge_Coverage) [![PyPI version](https://badge.fury.io/py/touchbar-lyric.svg)](https://badge.fury.io/py/touchbar-lyric)\n\nShow synced lyric in the touch-bar with BetterTouchTool and NetEase/QQ Music APIs. Based on the idea of [Kashi](https://community.folivora.ai/t/kashi-show-current-song-lyrics-on-touch-bar-spotify-itunes-youtube/6301).\n\n![Preview](./lyric_chinese.png)\n![Preview](./lyric_english.png)\n\n## Features\n\n-   **Synced lyrics** from QQ Music and NetEase Music APIs;\n-   Support **Spotify** (Recommended) & **Music** (Only songs in your playlists);\n-   Support for **English/Spanish/Chinese(Simplified/Traditional)/Japanese** and more;\n\n## Instruction\n\n**If you are not familiar with command line, python ecosystem or having problems understanding this tutorial, find a friend to help you. Issues/DMs are not actively monitored for this project.**\n\n### 1. Installation\n```shell\npip3 install touchbar_lyric --upgrade\n```\n\n### 2. Configuration in BetterTouchTool\n\nSame as Kashi:\n\n1.  Copy&paste the content in `lyric.json` in _Meun Bar > Touch Bar_;\n2.  Change the python path `$PYTHONPATH` to your own python path in the script area;\n\n```shell\n$PYTHONPATH -m touchbar_lyric --app Music\n```\n\nor use Spotify(default)\n\n```shell\n$PYTHONPATH -m touchbar_lyric --app Spotify\n```\n\nShow Traditional Chinese lyrics\n\n```shell\n$PYTHONPATH -m touchbar_lyric --app Spotify --traditional\n```\n\n**Be careful with typing double hyphens in BTT. It automatically change it to an em slash. Use copy & paste instead!**',
    'author': 'Chenghao Mou',
    'author_email': 'mouchenghao@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
